from .discord_types.member import Member
from .discord_types.user import User
from .discord_types.message import Message
from .incoming_interactions import (
    IncomingApplicationCommand,
    IncomingDiscordSlashInteraction,
    IncomingDiscordSlashData,
    SubcommandIncomingDiscordOptionList,
    IncomingDiscordOption,
    IncomingDiscordButtonInteraction,
    IncomingDiscordSelectMenuInteraction,
    SubcommandIncomingDiscordOptionListChild,
)



from enum import Enum
from pydantic import BaseModel, conlist
import typing
import dataclasses


if typing.TYPE_CHECKING:  # pragma: no cover
    static_check_init_args = dataclasses.dataclass
else:

    def static_check_init_args(cls):
        return cls


class AllowedMentionTypes(str, Enum):
    ROLE_MENTIONS = "roles"
    USER_MENTIONS = "users"
    EVERYONE_MENTIONS = "everyone"


@static_check_init_args
class AllowedMentions(BaseModel):
    parse: typing.List[AllowedMentionTypes]
    roles: conlist(str, max_items=100) = []
    users: conlist(str, max_items=100) = []
    replied_user: bool = False



from dispike.errors.events import InvalidEventType
import typing
from loguru import logger
import inspect
from enum import Enum


class EventTypes(str, Enum):
    COMMAND = "command"
    COMPONENT = "component"
    MESSAGE_COMMAND = "message_command"
    USER_COMMAND = "user_command"



from loguru import logger
import httpx


def dispike_httpx_event_hook_outgoing_request(request: httpx.Request):
    logger.opt(colors=True).log(
        "NETWORK",
        f"Outgoing request <yellow>[{request.method}]</yellow> <white>{request.url}</white>: headers: <white>{request.headers}</white>",
    )


def dispike_httpx_event_hook_incoming_request(response: httpx.Response):
    request = response.request  # type: httpx.Request
    logger.opt(colors=True).log(
        "NETWORK",
        f"Incoming request: <yellow>[{request.method}:{response.status_code}]</yellow> url: <white>{request.url}</white>: headers: <white>{request.headers}</white> ",
    )



import inspect
import typing
import warnings
from enum import Enum
from dispike.eventer import EventTypes
from fastapi import FastAPI
from loguru import logger

from dispike.errors.events import InvalidEventType
from dispike.errors.warnings import InsecureBindingWithCustomHostWarning
from dispike.errors.dispike import BotTokenNotProvided
from dispike.helper.network_request_log import (
    dispike_httpx_event_hook_incoming_request,
    dispike_httpx_event_hook_outgoing_request,
)
from .incoming import IncomingApplicationCommand
from .creating import RegisterCommands
from .creating.models import DiscordCommand
from .server import DiscordVerificationMiddleware
from .server import router
from .interactions import EventCollection, PerCommandRegistrationSettings

import asyncio

import httpx

from .errors.network import DiscordAPIError
from .creating.models.permissions import (
    GuildApplicationCommandPermissions,
    NewApplicationPermission,
)

if typing.TYPE_CHECKING:
    from .incoming import IncomingDiscordSlashInteraction  # pragma: no cover
    from .response import DiscordResponse  # pragma: no cover


class Dispike(object):
    """
    # Dispike
    An independent, simple to use, powerful framework for creating interaction-based Discord bots. 
    
    Powered by FastAPI
    """

    def __init__(
        self,
        client_public_key: str,
        application_id: str,
        bot_token: str = None,
        **kwargs,
    ):
        """Initialize Dispike Object

        Args:
            client_public_key (str): Discord provided client public key.
            bot_token (str): Discord provided bot token. Optional, but if you do not provide a bot token, you cannot register commands.
            application_id (str): Discord provided Client ID
            custom_context_argument_name (str, optional): Change the name of the context arugment when passing to a function. Set to "ctx".
            middleware_testing_skip_verification_key_request: (bool, optional): Skip middleware verification (NOT RECOMMENDED)
        """
        self._bot_token = bot_token
        self._application_id = application_id

        if bot_token is not None:
            self._registrator = RegisterCommands(
                application_id=self._application_id, bot_token=self._bot_token
            )
        else:
            self._registrator = False

        self._internal_application = FastAPI()

        if kwargs.get("middleware_testing_skip_verification_key_request", False):
            self._testing_skip_verification_key_request = True
        else:
            self._testing_skip_verification_key_request = False

        self._internal_application.add_middleware(
            DiscordVerificationMiddleware,
            client_public_key=client_public_key,
            testing_skip_verification_of_key=self._testing_skip_verification_key_request,
        )
        self._internal_application.include_router(router=router)
        if not kwargs.get("custom_context_argument_name"):
            router._user_defined_setting_ctx_value = "ctx"
        else:
            router._user_defined_setting_ctx_value = kwargs.get(
                "custom_context_argument_name"
            )
        self.callbacks = {
            "command": {},
            "component": {},
            "message_command": {},
            "user_command": {},
        }

        self._cache_router = router
        self._client = httpx.Client(
            base_url=f"https://discord.com/api/v8/applications/{self._application_id}/",
            event_hooks={
                "response": [dispike_httpx_event_hook_incoming_request],
                "request": [dispike_httpx_event_hook_outgoing_request],
            },
        )

    @logger.catch(reraise=True)
    def reset_registration(self, new_bot_token=None, new_application_id=None):
        """This method resets the built-in RgeisterCommands.
        You should not have to call this method directly.

        Call it only if you change the client id or bot token.

        Args:
            new_bot_token (None, optional): Description
            new_application_id (None, optional): Description

        Returns:
            TYPE: bool
        """
        if new_bot_token is None:
            _bot_token = self._bot_token
        else:
            _bot_token = new_bot_token
        if new_application_id is None:
            _application_id = self._application_id
        else:
            _application_id = new_application_id
        self._registrator = RegisterCommands(
            application_id=_application_id, bot_token=_bot_token
        )
        self._bot_token = _bot_token
        self._application_id = _application_id

        return True

    def return_bot_token_headers(self):
        if self._bot_token is not None:
            return {"Authorization": f"Bot {self._bot_token}"}
        raise BotTokenNotProvided("Authorizing Requests")

    @staticmethod
    async def background(function: typing.Callable, *args, **kwargs):
        logger.debug(f"register background to function {function}")
        return asyncio.create_task(function(*args, **kwargs))

    @property
    def referenced_application(self) -> FastAPI:
        """Returns the internal FastAPI object that was initialized.
        You are welcome to edit this with the appropriate settings found in
        the FastAPI docs.

        Returns:
            FastAPI: a pre-configured FastAPI object with required middlewares.
        """

        if router._dispike_instance == None:
            logger.debug(f"Adding self to _dispike_instance to router instance.")
            router._dispike_instance = self
        else:
            logger.debug("ApiRouter already has _dispike_instance that is not None.")
        return self._internal_application

    @property
    def register(self) -> RegisterCommands.register:
        """Returns a shortcut the RegisterCommands.register function

        Returns:
            RegisterCommands.register: internal RegisterCommands Object
        """
        if self._registrator == False:
            raise BotTokenNotProvided("Registrating commands")
        return self._registrator.register

    @property
    def shared_client(self) -> "httpx.Client":
        """Returns a pre-initialized ``httpx.Client`` that is used for requests internally.

        Returns:
            httpx.Client: used for network requests to discord.
        """
        return self._client

    @property
    def interaction(self):
        """
        # This is deprecated and will be removed in a future version of dispike!
        use `self.on` instead.

        This method returns ``self.on`` to ensure compatiablity with the previous versions of dispike.
        """

        warnings.warn(
            "Using .interaction directly is deprecated and will be removed in future versions of Dispike. Please use the .on method on your dispike object instead.",
            DeprecationWarning,
        )  # pragma: no cover
        return self.on

    @logger.catch(reraise=True, message="Issue with getting commands from Discord")
    def get_commands(
        self, guild_only=False, guild_id_passed=None
    ) -> typing.List[IncomingApplicationCommand]:
        """Returns a list of ``DiscordCommands`` either globally or for a specific guild.

        Args:
            guild_only (bool, optional): whether to target a guild. Defaults to False.
            guild_id_passed ([type], optional): guild id if guild_only is set to True. Defaults to None.

        Returns:
            typing.List[DiscordCommand]: Array of DiscordCommand

        Raises:
            DiscordAPIError: any Discord returned errors.
        """
        if guild_only:
            if not guild_id_passed or not isinstance(guild_id_passed, str):
                raise TypeError(
                    "You cannot have guild_only == True and NOT pass any guild id."
                )
            _url = f"/guilds/{guild_id_passed}/commands"
        else:
            _url = f"/commands"

        try:
            _send_request = self._registrator._client.get(
                _url, headers=self._registrator.request_headers
            )
            if _send_request.status_code == 200:
                return [IncomingApplicationCommand(**x) for x in _send_request.json()]

            raise DiscordAPIError(_send_request.status_code, _send_request.text)
        except DiscordAPIError:
            logger.exception("Discord API Failure.")
            raise

    @logger.catch(reraise=True, message="Issue with editing commands from Discord")
    def edit_command(
        self,
        new_command: typing.Union[typing.List[DiscordCommand], DiscordCommand],
        command_id: typing.Union[int, IncomingApplicationCommand] = None,
        bulk=False,
        guild_only=False,
        guild_id_passed=False,
    ) -> DiscordCommand:
        """Edits a command provided with a command_id and a valid new command.

        Args:
            command_id (int): Command ID
            new_command ([DiscordCommand, List[DiscordCommand]]): A valid DiscordCommand object (or a dict with proper syntax, if a dict is passed no verification will be made and discord will return the syntax error)
            guild_only (bool, optional): whether to target a guild. Defaults to False.
            guild_id_passed (bool, optional): guild id if guild_only is set to True. Defaults to None.
            bulk (bool, optional): Whether to specifiy if this action will be a bulk action.

        Returns:
            DiscordCommand: Returns the DiscordCommand object created. (Will return a DiscordCommand irregardless of new_command)

        Raises:
            TypeError: Invalid types passed.
            DiscordAPIError: any Discord returned errors.
        """

    
        if command_id:
            if isinstance(command_id, IncomingApplicationCommand):
                command_id = command_id.id
            elif isinstance(command_id, (str, int)):
                command_id = int(command_id)
            else:
                raise TypeError("The command ID must be either an interger or an IncomingApplicationCommand object.")
        
        if not isinstance(new_command, (DiscordCommand, dict, list)):
            raise TypeError("New command must be a DiscordCommand or a valid dict.")

        if guild_only:
            if not guild_id_passed:
                raise TypeError(
                    "You cannot have guild_only set to True and NOT pass any guild id."
                )
            if bulk:
                _url = f"/guilds/{guild_id_passed}/commands"
            else:
                _url = f"/guilds/{guild_id_passed}/commands/{command_id}"
        else:
            _url = "/commands"
        if bulk == True and isinstance(new_command, list):
            _new_command = [command.dict() for command in new_command]
            _selected_request_method = "PUT"
        else:
            _new_command = new_command.dict()
            _selected_request_method = "PATCH"
        try:
            _send_request = self._registrator._client.request(
                method=_selected_request_method,
                url=_url,
                headers=self._registrator.request_headers,
                json=_new_command,
            )
            if _send_request.status_code != 200:
                raise DiscordAPIError(_send_request.status_code, _send_request.text)

            if bulk:
                return [DiscordCommand(**x) for x in _send_request.json()]
            else:
                return DiscordCommand(**_send_request.json())
        except DiscordAPIError:
            # TODO: Maybe don't return false and just raise it?
            logger.exception("Discord API Failure.")
            return False

    @logger.catch(reraise=True, message="Issue with deleting commands from Discord")
    def delete_command(
        self, command_id: typing.Union[int, IncomingApplicationCommand], guild_only=False, guild_id_passed=None
    ) -> bool:
        """Deletes a command, provided with a command_id

        Args:
            command_id (typing.Union[int, IncomingApplicationCommand]): Command ID required
            guild_only (bool, optional): Whether to be a global action or target a guild. Defaults to False.
            guild_id_passed ([type], optional): Guild ID if guild_only is set to True. Defaults to None.

        Returns:
            bool: True if status code is 201, otherwise an error will be raised.

        Raises:
            TypeError: Invalid types passed.
            DiscordAPIError: any Discord returned errors.
        """

        if isinstance(command_id, IncomingApplicationCommand):
            command_id = command_id.id
        elif isinstance(command_id, (str, int)):
            command_id = int(command_id)
        else:
            raise TypeError("The command ID must be either an interger or an IncomingApplicationCommand object.")
        
        if guild_only:
            if not guild_id_passed:
                raise TypeError(
                    "You cannot have guild_only == True and NOT pass any guild id."
                )
            _url = f"/guilds/{guild_id_passed}/commands/{command_id}"
        else:
            _url = f"/commands/{command_id}"

        try:
            _send_request = self._registrator._client.delete(
                _url, headers=self._registrator.request_headers
            )
            if _send_request.status_code != 204:
                raise DiscordAPIError(_send_request.status_code, _send_request.text)
            return True
        except DiscordAPIError:
            logger.exception("Discord API Failure.")
            raise

    def set_command_permission(
        self, command_id: typing.Union[int, IncomingApplicationCommand], guild_id: int, new_permissions: "NewApplicationPermission"
    ) -> bool:
        """Set a permissions for a command in a specific guild. This function is sync!

        Args:
            command_id (typing.Union[int, IncomingApplicationCommand]): Either a command id int or a IncomingApplicationCommand (obtained from .get_command)
            guild_id (int): The guild to be targeted.
            new_permissions (NewApplicationPermission): Permissions for this command.

        Returns:
            bool: True, if the command has been successfully edited.
        """
        
        if isinstance(command_id, IncomingApplicationCommand):
            command_id = command_id.id
        elif isinstance(command_id, (str, int)):
            command_id = int(command_id)
        else:
            raise TypeError("The command ID must be either an interger or an IncomingApplicationCommand object.")
        
        with httpx.Client() as client:
            try:

                _set_command_permissions = client.put(
                    f"https://discord.com/api/v8/applications/{self._application_id}/guilds/{guild_id}/commands/{command_id}/permissions",
                    json=new_permissions.dict(),
                    headers=self.return_bot_token_headers(),
                )
                _set_command_permissions.raise_for_status()
                return True
            except httpx.HTTPError:
                logger.exception(
                    f"Unable to set permission for command {command_id} for guild {guild_id}"
                )
                logger.debug(
                    f"request: {_set_command_permissions.status_code}: {_set_command_permissions.text}"
                )
                return False

    async def async_set_command_permission(
        self, command_id: typing.Union[int, IncomingApplicationCommand], guild_id, new_permissions: "NewApplicationPermission"
    ) -> bool:
        """Set permissions for a command in a specific guild. This function is async!

        Args:
            command_id (typing.Union[int, IncomingApplicationCommand]): Either a command id int or a IncomingApplicationCommand (obtained from .get_command)
            guild_id (int): The guild to be targeted.
            new_permissions (NewApplicationPermission): Permissions for this command.

        Returns:
            bool: True, if the command has been successfully edited.
        """        

        if isinstance(command_id, IncomingApplicationCommand):
            command_id = command_id.id
        elif isinstance(command_id, (str, int)):
            command_id = int(command_id)
        else:
            raise TypeError("The command ID must be either an interger or an IncomingApplicationCommand object.")
        
        async with httpx.AsyncClient() as client:
            try:

                _set_command_permissions = await client.put(
                    f"https://discord.com/api/v8/applications/{self._application_id}/guilds/{guild_id}/commands/{command_id}/permissions",
                    data=new_permissions.dict(),
                    headers=self.return_bot_token_headers(),
                )
                _set_command_permissions.raise_for_status()
                return True
            except httpx.HTTPError:
                logger.exception(
                    f"Unable to set permission for command {command_id} for guild {guild_id}"
                )
                return False

    async def send_deferred_message(
        self,
        original_context: "IncomingDiscordSlashInteraction",
        new_message: "DiscordResponse",
    ):
        """Send a deferred message.

        Args:
            original_context (IncomingDiscordSlashInteraction): The orginal context of the message.
            new_message (DiscordResponse): Message to send.
        """
        async with httpx.AsyncClient(
            base_url=f"https://discord.com/api/v8/webhooks/{self._application_id}/{original_context.token}/messages/",
            headers={
                "Content-Type": "application/json",
            },
        ) as client:
            try:
                # TODO: Probably change later to inside the DeferredResponse?
                new_message._switch_to_followup_message()
                logger.debug(f"sending deferred response : {new_message.response}")
                response = await client.patch("/@original", json=new_message.response)
                response.raise_for_status()
            except httpx.HTTPError as req:
                logger.exception(f"Unable to send deferred message!")
                raise

    async def async_get_command_permission_in_guild(
        self, command_id, guild_id
    ) -> GuildApplicationCommandPermissions:
        """Return permissions for a single command in a guild. If no permissions are available, it will return None.

        Args:
            command_id (typing.Union[str, int]): Command ID
            guild_id (typing.Union[str, int]): Guild ID

        Returns:
            GuildApplicationCommandPermissions: Return if permissions exist.
            None: Return if no permissions exist.
        """
        async with httpx.AsyncClient() as client:
            try:
                _request_command_permission = await client.get(
                    f"https://discord.com/api/v8/applications/{self._application_id}/guilds/{guild_id}/commands/{command_id}/permissions"
                )
                if _request_command_permission.status_code == 404:
                    return None
                elif _request_command_permission.status_code == 200:
                    return GuildApplicationCommandPermissions(
                        **_request_command_permission.json()
                    )
                else:
                    raise DiscordAPIError(
                        status_code=_request_command_permission.status_code,
                        request_text=_request_command_permission.text,
                    )
            except DiscordAPIError:
                logger.error(
                    f"Unable to get command permission! {_request_command_permission.status_code}"
                )
                raise

    async def async_get_all_command_permissions_in_guild(
        self, guild_id
    ) -> typing.List[GuildApplicationCommandPermissions]:
        """Return permissions for all commands in a guild.

        Args:
            guild_id (typing.Union[str, int]): ID of guild.

        Returns:
            typing.List[GuildApplicationCommandPermissions]: Permissions for all commands (if any permissions exist.)
        """
        async with httpx.AsyncClient() as client:
            try:
                _request_command_permission = await client.get(
                    f"https://discord.com/api/v8/applications/{self._application_id}/guilds/{guild_id}/commands/permissions"
                )

                if _request_command_permission.status_code not in [200, 201]:
                    raise DiscordAPIError(
                        status_code=_request_command_permission.status_code,
                        request_text=_request_command_permission.text,
                    )
                return [
                    GuildApplicationCommandPermissions(**x)
                    for x in _request_command_permission.json()
                ]
            except DiscordAPIError:
                raise

    def get_all_command_permissions_in_guild(
        self, guild_id: typing.Union[str, int]
    ) -> typing.List[GuildApplicationCommandPermissions]:
        """Return permissions for all commands in a guild.

        Args:
            guild_id (typing.Union[str, int]): ID of guild.

        Returns:
            typing.List[GuildApplicationCommandPermissions]: Permissions for all commands (if any permissions exist.)
        """
        with httpx.Client() as client:
            try:
                _request_command_permission = client.get(
                    f"https://discord.com/api/v8/applications/{self._application_id}/guilds/{guild_id}/commands/permissions",
                    headers=self.return_bot_token_headers(),
                )
                if _request_command_permission.status_code not in [200, 201]:
                    raise DiscordAPIError(
                        status_code=_request_command_permission.status_code,
                        request_text=_request_command_permission.text,
                    )
                return [
                    GuildApplicationCommandPermissions(**x)
                    for x in _request_command_permission.json()
                ]
            except DiscordAPIError:
                raise

    def get_command_permission_in_guild(
        self, command_id: typing.Union[str, int], guild_id: typing.Union[str, int]
    ) -> GuildApplicationCommandPermissions:
        """Return permissions for a single command in a guild. If no permissions are available, it will return None.

        Args:
            command_id (typing.Union[str, int]): Command ID
            guild_id (typing.Union[str, int]): Guild ID

        Returns:
            GuildApplicationCommandPermissions: Return if permissions exist.
            None: Return if no permissions exist.
        """
        with httpx.Client() as client:
            try:
                _request_command_permission = client.get(
                    f"https://discord.com/api/v8/applications/{self._application_id}/guilds/{guild_id}/commands/{command_id}/permissions",
                    headers=self.return_bot_token_headers(),
                )
                if _request_command_permission.status_code == 404:
                    return None
                elif _request_command_permission.status_code == 200:
                    return GuildApplicationCommandPermissions(
                        **_request_command_permission.json()
                    )
                else:
                    raise DiscordAPIError(
                        status_code=_request_command_permission.status_code,
                        request_text=_request_command_permission.text,
                    )
            except DiscordAPIError:
                logger.error(
                    f"Unable to get command permission! {_request_command_permission.status_code}"
                )
                raise

    @staticmethod
    def _return_uvicorn_run_function():
        """Import uvicorn, only exists to make testing easier. You do not need to import this.

        Raises:
            SystemExit: If uvicorn is not installed

        Returns:
            uvicorn: If everything works out.
        """
        try:
            import uvicorn

            return uvicorn
        except Exception:
            raise SystemExit(
                "Uvicorn is not installed. Please use a different webserver pointing to <..>.referenced_application"
            )

    def run(
        self,
        port: int = None,
        unix_socket: str = None,
        bind_to_ip_address: str = None,
        supress_insecure_binding_warning: bool = False,
    ):
        """Runs the bot with the already-installed Uvicorn webserver.

        Args:
            port (int, optional): Run the bot on a specific port.
            unix_socket (str, optional): [description]. Run the bot and listen on a specific unix domain socket..

        Raises:
            ArgumentError: Raised if you pass both a port and a unix socket.
        """
        uvicorn = self._return_uvicorn_run_function()

        if unix_socket and port:
            raise ValueError("You cannot bind to port AND a unix socket")
        else:
            if port:
                if bind_to_ip_address:
                    if supress_insecure_binding_warning == False:
                        warnings.warn(
                            "Binding to a IP Address other than 127.0.0.1 may not be secure! If you are exposing this service to the outside world -- a reverse proxy is strongly recommended.",
                            InsecureBindingWithCustomHostWarning,
                        )
                    uvicorn.run(
                        app=self.referenced_application,
                        host=bind_to_ip_address,
                        port=port,
                    )
                else:
                    uvicorn.run(app=self.referenced_application, port=port)
            elif unix_socket:
                if "unix" not in unix_socket:
                    unix_socket = f"unix:{unix_socket}"
                else:
                    uvicorn.run(self.referenced_application, host=unix_socket)
            if not unix_socket and not port:
                raise ValueError("You must specify a port or unix socket")

    def on(
        self,
        event: str,
        type: EventTypes = EventTypes.COMMAND,
        func: typing.Callable = None,
    ):
        """A wrapper over an async function, registers it in .callbacks.

        Args:
            event (str): Event name
            type (EventTypes): Type of this event.
            func (None, optional): function to wrap around

        Returns:
            <function>: returns the wrapped function
        """

        if isinstance(type, (EventTypes, str)):

            # TODO: This is broken, the only way we can show this warning is by
            # checking if the value is an Enum, which requires builtin-type (which we overriden)
            """
            if not isinstance(type, EventTypes):  # pragma: no cover

                logger.warning(
                    "Passing a unknown EventType, this may cause issues and is unsupported"
                )  # noqa
            else:
                # TODO: Maybe it's not good to overrwrite a default python function. Maybe change type to a different value?
                raise InvalidEventType(type)  # pragma: no cover
            """
        else:
            raise InvalidEventType(type)

        def on(func):
            self._add_function_to_callbacks(event, type, func)
            return func

        return on(func) if func else on

    def check_event_exists(self, event: str, type: str) -> bool:
        """Checks if the event in ``.callbacks``

        Args:
            event (str): event name

        Returns:
            bool: returns if the event is in callbacks.
        """
        return event in self.callbacks[type]

    def return_event_settings(self, event: str, type: str) -> dict:

        if self.check_event_exists(event, type):
            return self.callbacks[type][event]["settings"]
        raise TypeError(
            f"Event {event} is not in callbacks. Did you register this event?"
        )

    def return_event_function(self, event: str, type: str) -> dict:
        """Returns the function registered for the event.

        Args:
            event (str): Event name
            type (str): Event typelogger.debug(f"logMessage")

        Raises:
            TypeError: If Event name is not in callbacks.

        Returns:
            dict: containing func and settings
        """
        if self.check_event_exists(event, type):
            return self.callbacks[type][event]["function"]
        raise TypeError(
            f"Event {event} is not in callbacks. Did you register this event?"
        )

    def view_event_function_return_type(self, event: str, type: str) -> dict:
        """Get type hint for event functions

        Args:
            event (str): Event name

        Returns:
            dict: Returns .get_type_hints for event
        """
        return typing.get_type_hints(self.callbacks[type][event]["function"])

    async def emit(self, event: str, type: str, *args, **kwargs):
        """'Emits' an event. It will basically call the function from .callbacks and return the function result

        Args:
            event (str): Event name
            type (str): Event type
            *args: extra arguments to pass
            **kwargs: extra kwargs to pass

        Returns:
            function result: returns the function result

        Raises:
            TypeError: raises if event is not registered.
        """
        if event not in self.callbacks[type]:
            raise TypeError(
                f"event {event} does not have a corresponding handler. Did you register this function/event?"
            )

        _look_up_function = self.return_event_function(event, type)
        return await _look_up_function(*args, **kwargs)

    def _add_function_to_callbacks(
        self, function_name: str, function_type: EventTypes, function: typing.Callable
    ):
        """Internal function, adds a function to the callbacks.

        Args:
            function_name (str): Function event name, this will be used to look up the function in .callbacks
            function_type (EventTypes): Function type
            function (typing.Callable): Actual function, should be an async function.

        Raises:
            TypeError: If function is not an async function.
            TypeError: If function name has already been registered
            TypeError: Invalid EventType
        """
        if not inspect.iscoroutinefunction(function):
            raise TypeError("Passed function is not an asynchronous function.")

        if not isinstance(function_type, (EventTypes, str)):
            raise TypeError(
                f"Passed function is not the correct type. Expected a <str> but received {type(function_type)}"
            )

        if function_name in self.callbacks[function_type]:
            raise TypeError(f"{function_name} ({function_type}) is already registered.")
        else:
            logger.debug(f"Adding {function_name} ({function_type}) to callbacks..")
            self.callbacks[function_type][function_name] = {
                "settings": {},
                "function": function,
            }

    def register_event_command(
        self,
        function_event_name: str,
        function: typing.Callable,
        function_type: EventTypes = None,
        **kwargs,
    ):
        """This will register an event command.
        Usually you would use this if you are using dispike.interactions.on; instead of the self.on.

        Args:
            function_event_name (str): Function event name
            function (typing.Callable): Function to register. Must be async
            function_type (EventTypes): [description]. Function type, defaults to None, but will still attempt to find the type. Will raise an error if it cannot.

        Raises:
            AttributeError: If function_type is not a string or None, and it cannot be found in the functions attributes.
        """
        if function_type is None:
            # try to see if the function has a _event_type attribute
            try:
                function_type = function._dispike_event_type
            except AttributeError:
                raise AttributeError(
                    "Unable to find function event type attribute inside function.. Did you add a decorator to this function?"
                )
        else:
            if isinstance(function_type, EventTypes):
                function_type = function_type

        self._add_function_to_callbacks(
            function_name=function_event_name,
            function_type=function_type,
            function=function,
        )

    def register_collection(
        self,
        collection: "EventCollection",
        register_command_with_discord: bool = False,
        initialze_on_load: bool = False,
        initialization_arguments: typing.Dict = None,
    ):
        """Registers a EventCollection.

        Args:
            collection (EventCollection): The collection to register.
        """

        if initialze_on_load:
            collection = collection(**initialization_arguments)

        _load_in_functions = self._detect_functions_with_event_decorator(
            collection=collection,
        )

        # TODO: Maybe re-enable this as a fallback?
        """
        for shallow_function in collections.registered_commands():
            self._add_function_to_callbacks(
                function=shallow_function,
                function_name=shallow_function._dispike_event_name,
                function_type=shallow_function._dispike_event_type,
            )
        """

        for shallow_function in _load_in_functions:
            self._add_function_to_callbacks(
                function=shallow_function,
                function_name=shallow_function._dispike_event_name,
                function_type=shallow_function._dispike_event_type,
            )

        if register_command_with_discord:
            for command in collection.command_schemas():
                if isinstance(command, PerCommandRegistrationSettings):
                    self.register(
                        command=command.schema,
                        guild_only=True,
                        guild_to_target=command.guild_id,
                    )
                else:
                    self.register(command=command)

    def clear_all_event_callbacks(self, event_type: "EventTypes" = None):
        """Clears all event callbacks."""
        if not event_type:
            self.callbacks = {"command": {}, "component": {}}
        else:
            self.callbacks[event_type] = {}

    @staticmethod
    def _detect_functions_with_event_decorator(collection: EventCollection):
        """Internal function, detects functions with the event decorator inside a EventCollection.

        Args:
            collection (EventCollection): Event collection contatining at least one function with the interactions decorator.

        Raises:
            ValueError: If the EventCollection contains no functions with the interactions decorator.
            TypeError: If the collection passed is not a EventCollection or subclass of it.

        Returns:
            [type]: [description]
        """
        # check if colleciton inherited from EventCollection
        if isinstance(collection, EventCollection) or issubclass(
            collection, EventCollection
        ):

            _funcs = []
            for _functions in inspect.getmembers(
                collection, inspect.iscoroutinefunction
            ):
                if hasattr(_functions[1], "_dispike_event_name"):
                    _funcs.append(_functions[1])

            if len(_funcs) == 0:
                raise ValueError(
                    f"{collection} contains no functions that have the interactions decorator."
                )
            return _funcs
        raise TypeError("Collection passed must subclass EventCollection.")



from dispike.errors.response import NoResolvedInteractions
from .attribute_helpers import (
    lookup_resolved_member_helper,
    lookup_resolved_user_helper,
    lookup_resolved_channel_helper,
    lookup_resolved_role_helper,
)
from pydantic import BaseModel, ValidationError, validator
import typing
from .discord_types.member import Member, PartialMember
from .discord_types.message import Message
from .discord_types.user import User
from .discord_types.role import Role
from .discord_types.channel import PartialChannel
from loguru import logger

from ..creating import CommandOption, SubcommandOption
from ..creating.models.options import CommandTypes

try:
    from typing import Literal  # pragma: no cover
except ImportError:  # pragma: no cover
    # backport
    from typing_extensions import Literal  # pragma: no cover


class IncomingDiscordOption(BaseModel):

    """An incoming discord option, this is not intended for you to edit, and will not
    be accepted as an argument in any function nor will be accepted as a value in DiscordCommand

    Attributes:
        name (str): Name of the option
        value (str): Value of the option
    """

    name: str
    value: str


class SubcommandIncomingDiscordOptionListChild(BaseModel):

    """
    Attributes:
        name (str): Name of the subcommand
        options (List[IncomingDiscordOption]], optional): Options selected by the user
    """

    class Config:
        arbitary_types_allowed = True

    name: str
    options: typing.List[IncomingDiscordOption]


class SubcommandIncomingDiscordOptionList(BaseModel):

    """An incoming discord option list, this is not intended for you to edit, and will not
    be accepted as an argument in any function nor accepted in DiscordCommand

    Attributes:
        name (str): Name of the subcommand
        options (Union[List[IncomingDiscordOption], List[SubcommandIncomingDiscordOptionListChild]], optional): Options selected by the user
    """

    class Config:
        arbitary_types_allowed = True

    name: str
    options: typing.Union[
        typing.List[IncomingDiscordOption],
        typing.List[SubcommandIncomingDiscordOptionListChild],
    ]
    resolved: typing.Optional[typing.Dict] = {}


class IncomingDiscordSlashData(BaseModel):

    """An incoming discord option list, this is not intended for you to edit, and will not
    be accepted as an argument in any function nor accepted in DiscordCommand

    Attributes:
        id (str): Id of the command
        name (str): Name of the command
        options (Union[List[IncomingDiscordOption], List[SubcommandIncomingDiscordOptionList]], optional): Options selected by the user
    """

    class Config:
        arbitary_types_allowed = True

    id: str
    name: str
    options: typing.Optional[
        typing.Union[
            typing.List[IncomingDiscordOption],
            typing.List[SubcommandIncomingDiscordOptionList],
        ]
    ] = None
    resolved: typing.Optional[typing.Dict] = {}
    type: CommandTypes = CommandTypes.SLASH


class IncomingDiscordButtonData(BaseModel):

    """
    Incoming button data.

    Attributes:
        custom_id (str): The custom id of this button.
    """

    custom_id: str


class IncomingDiscordSelectMenuData(BaseModel):

    """
    Incoming select menu data.

    Attributes:
        custom_id (str): The custom id of this button.
        values (List[str]): A list of string values that the user selected.
    """

    custom_id: str
    values: typing.List[str]


class IncomingDiscordUserCommandData(BaseModel):

    """
    Incoming user command data.

    Attributes:
        target_id (str): The id of the targeted user.
        target (Member): The targeted member.
        name (str): The name of the user command.
    """

    target_id: str
    target: Member = None
    name: str


class IncomingDiscordMessageCommandData(BaseModel):

    """
    Incoming message command data.

    Attributes:
        target_id (str): The id of the targeted message.
        target (Message): The target message.
        name (str): The name of the message command.
    """

    target_id: str
    target: Message = None
    name: str


class IncomingDiscordSlashInteraction(BaseModel):

    """An incoming discord interaction that was triggered by a command, this is not intended for you to edit, and will not
    be accepted as an argument in any function.

    Attributes:
        id (int): Id of the interaction.
        data (IncomingDiscordSlashData): Options from the command.
        guild_id (int): Guild ID where this happened.
        channel_id (int): Channel ID where this happened.
        member (Member): Member that used this interaction.
        token (str): Token of this interaction.
    """

    class Config:
        arbitary_types_allowed = True

    type: Literal[2, 3, 4, 5, 6, 7, 8]  # 1 is removed, this lib will handle PING
    id: int
    data: IncomingDiscordSlashData
    guild_id: int
    channel_id: int
    member: Member
    token: str
    version: typing.Optional[Literal[1]] = None

    def lookup_resolved_member(
        self: "IncomingDiscordSlashInteraction", member_id: str
    ) -> "PartialMember":
        if self.data.resolved != {}:
            return lookup_resolved_member_helper(cls=self, member_id=member_id)
        else:
            raise NoResolvedInteractions

    def lookup_resolved_user(
        self: "IncomingDiscordSlashInteraction", user_id: str
    ) -> "User":
        if self.data.resolved != {}:
            return lookup_resolved_user_helper(cls=self, user_id=user_id)
        else:
            raise NoResolvedInteractions

    def lookup_resolved_channel(
        self: "IncomingDiscordSlashInteraction", channel_id: str
    ) -> "PartialChannel":
        if self.data.resolved != {}:
            return lookup_resolved_channel_helper(cls=self, channel_id=channel_id)
        else:
            raise NoResolvedInteractions

    def lookup_resolved_role(
        self: "IncomingDiscordSlashInteraction", role_id: str
    ) -> "Role":
        if self.data.resolved != {}:
            return lookup_resolved_role_helper(cls=self, role_id=role_id)
        else:
            raise NoResolvedInteractions


class IncomingDiscordButtonInteraction(BaseModel):

    """An incoming discord interaction that was triggered by a button press, this is not intended for you to edit, and will not
    be accepted as an argument in any function.

    Attributes:
        id (int): Id of the interaction.
        data (IncomingDiscordButtonData): Data from the interaction.
        guild_id (int): Guild ID where this happened.
        channel_id (int): Channel ID where this happened.
        member (Member): TODO: IDK if this is the member that called the command originally or the one that used the button.
        token (str): Token of this interaction.
    """

    type: Literal[2, 3, 4, 5, 6, 7, 8]  # 1 is removed, this lib will handle PING
    id: int
    data: IncomingDiscordButtonData
    guild_id: int
    channel_id: int
    member: Member
    token: str
    version: typing.Optional[Literal[1]] = None


class IncomingDiscordSelectMenuInteraction(BaseModel):

    """An incoming discord interaction that was triggered by a select menu interaction, this is not intended for you to edit, and will not
    be accepted as an argument in any function.

    Attributes:
        id (int): Id of the interaction.
        data (IncomingDiscordSelectMenuData): Data from the interaction.
        guild_id (int): Guild ID where this happened.
        channel_id (int): Channel ID where this happened.
        member (Member): TODO: IDK if this is the member that called the command originally or the one that used the select menu.
        token (str): Token of this interaction.
    """

    type: Literal[2, 3, 4, 5, 6, 7, 8]  # 1 is removed, this lib will handle PING
    id: int
    data: IncomingDiscordSelectMenuData
    guild_id: int
    channel_id: int
    member: Member
    token: str
    version: typing.Optional[Literal[1]] = None


class IncomingDiscordUserCommandInteraction(BaseModel):

    """An incoming discord interaction that was triggered by a user command interaction, this is not intended for you to edit, and will not
    be accepted as an argument in any function.

    Attributes:
        id (int): Id of the interaction.
        data (IncomingDiscordUserCommandData): Data from the interaction.
        guild_id (int): Guild ID where this happened.
        channel_id (int): Channel ID where this happened.
        member (Member): Member that called the command.
        token (str): Token of this interaction.
    """

    type: Literal[2, 3, 4, 5, 6, 7, 8]  # 1 is removed, this lib will handle PING
    id: int
    data: IncomingDiscordUserCommandData
    guild_id: int
    channel_id: int
    member: Member
    token: str
    version: typing.Optional[Literal[1]] = None


class IncomingDiscordMessageCommandInteraction(BaseModel):

    """An incoming discord interaction that was triggered by a message command interaction, this is not intended for you to edit, and will not
    be accepted as an argument in any function.

    Attributes:
        id (int): Id of the interaction.
        data (IncomingDiscordMessageCommandData): Data from the interaction.
        guild_id (int): Guild ID where this happened.
        channel_id (int): Channel ID where this happened.
        member (Member): Member that called the command.
        token (str): Token of this interaction.
    """

    type: Literal[2, 3, 4, 5, 6, 7, 8]  # 1 is removed, this lib will handle PING
    id: int
    data: IncomingDiscordMessageCommandData
    guild_id: int
    channel_id: int
    member: Member
    token: str
    version: typing.Optional[Literal[1]] = None


class IncomingApplicationCommand(BaseModel):

    """An Incoming Application command, this is not intended for you to edit, and will not be accepted
    in any function

    Attributes:
        id (int): Id of the interaction.
        application_id (int): The id of your bot.
        name (int): Name of the command.
        description (int): Description of the command.
        options (Union[List[CommandOption], List[SubcommandOption]], optional): Selected options from the command.
        default_permission (bool, optional): Bool whether if uses default permissions.
    """

    class Config:
        arbitary_types_allowed = True

    id: int
    application_id: int
    name: str
    description: str
    options: typing.Optional[
        typing.Union[typing.List[CommandOption], typing.List[SubcommandOption]]
    ]
    default_permission: typing.Optional[bool]

    # ? not listed in docs but appears in request
    version: typing.Optional[str]



class DiscordAPIError(Exception):
    # TODO: Add detection for more distinct exceptions based on api codes.
    def __init__(self, status_code, request_text):
        self.message = f"Discord API returned an unexpected response back [{status_code}]: {request_text}"

    def __str__(self):
        return self.message



from .registrator import RegisterCommands
from .models import DiscordCommand, SubcommandOption, CommandChoice, CommandOption
from .components import Button, LinkButton, ButtonStyles
from .allowed_mentions import AllowedMentions, AllowedMentionTypes


from fastapi import APIRouter, Request, Response
from fastapi.responses import PlainTextResponse
from loguru import logger

from .incoming.incoming_interactions import IncomingDiscordUserCommandInteraction, \
    IncomingDiscordMessageCommandInteraction
from .middlewares.verification import DiscordVerificationMiddleware
from dispike.incoming import (
    IncomingDiscordSlashInteraction,
    IncomingDiscordSlashData,
    SubcommandIncomingDiscordOptionList,
    IncomingDiscordOption,
    IncomingDiscordButtonInteraction,
    IncomingDiscordSelectMenuInteraction, Member, Message
)
from .eventer import EventTypes
from .eventer_helpers.determine_event_information import determine_event_information
from .response import DiscordResponse, DeferredResponse, DeferredEmphericalResponse
from dispike.creating.components import ComponentTypes
import json
import typing
import asyncio
import warnings


if typing.TYPE_CHECKING:
    from .main import Dispike

router = APIRouter()
router._dispike_instance = None
interaction = router._dispike_instance  # type: Dispike


_RAISE_FOR_TESTING = False


async def _run_and_log_async(coroutine: typing.Coroutine) -> None:
    logger.debug(f"Incoming deferred coroutine.. {coroutine}")
    await coroutine
    logger.debug(f"Deferred coroutine completed!")


@router.get("/ping")
async def ping():
    return PlainTextResponse(
        "If you see this, Your instance is working and accepting requests."
    )


@router.post("/interactions")
async def handle_interactions(request: Request) -> Response:
    if router._dispike_instance == None:
        return PlainTextResponse(
            "If you see this, Dispike has not been properly configured. Make sure to create a Dispike instance before starting this endpoint."
        )

    logger.info("interaction recieved.")

    _get_request_body = json.loads(request.state._cached_body.decode())
    logger.info(_get_request_body)
    if _get_request_body["type"] == 1:
        logger.info("handling ACK Ping.")
        return {"type": 1}

    if _get_request_body["type"] == 2:
        # ["data"]["type"] represents if its a message or user command

        if _get_request_body["data"]["type"] == 2:
            # 2 is user command

            _parse_to_object = IncomingDiscordUserCommandInteraction(
                **_get_request_body
            )

            # Confusingly construct a member object
            _member = Member(
                **{
                    **_get_request_body["data"]["resolved"]["members"][
                        _parse_to_object.data.target_id
                    ],
                    "user": {
                        **_get_request_body["data"]["resolved"]["users"][
                            _parse_to_object.data.target_id
                        ]
                    },
                }
            )

            # Set that member object as the target
            _parse_to_object.data.__setattr__("target", _member)

            _get_res = await router._dispike_instance.emit(
                _get_request_body["data"]["name"],
                EventTypes.USER_COMMAND,
                _parse_to_object,
            )
            return _get_res.response
        elif _get_request_body["data"]["type"] == 3:
            # 3 is message command

            _parse_to_object = IncomingDiscordMessageCommandInteraction(
                **_get_request_body
            )

            _message = Message(
                **_get_request_body["data"]["resolved"]["messages"][
                    _parse_to_object.data.target_id
                ]
            )

            # Set that message object as the target
            _parse_to_object.data.__setattr__("target", _message)

            _get_res = await router._dispike_instance.emit(
                _get_request_body["data"]["name"],
                EventTypes.MESSAGE_COMMAND,
                _parse_to_object,
            )
            return _get_res.response

    if _get_request_body["type"] == 3:
        if _get_request_body["data"]["component_type"] == ComponentTypes.BUTTON.value:
            # Button
            _parse_to_object = IncomingDiscordButtonInteraction(**_get_request_body)

            _get_res = await router._dispike_instance.emit(
                _get_request_body["data"]["custom_id"],
                EventTypes.COMPONENT,
                _parse_to_object,
            )
            return _get_res.response
        if (
            _get_request_body["data"]["component_type"]
            == ComponentTypes.SELECT_MENU.value
        ):
            # Select Menu
            _parse_to_object = IncomingDiscordSelectMenuInteraction(**_get_request_body)

            _get_res = await router._dispike_instance.emit(
                _get_request_body["data"]["custom_id"],
                EventTypes.COMPONENT,
                _parse_to_object,
            )
            return _get_res.response

    _parse_to_object = IncomingDiscordSlashInteraction(**_get_request_body)
    _event_name, arguments = determine_event_information(_parse_to_object)
    logger.debug(f"incoming event name: {_event_name}")
    if not router._dispike_instance.check_event_exists(_event_name, EventTypes.COMMAND):
        logger.debug("discarding event not existing.")
        warnings.warn(
            f"Event {_event_name} does not exist or does not have a callback.",
            UserWarning,
        )
        return {"type": 5}

    # _event_settings = router._dispike_instance.return_event_settings(_event_name)

    arguments[router._user_defined_setting_ctx_value] = _parse_to_object

    # Check the type hint for the return type, fallback for checking the type if no hints are provided
    try:
        _type_hinted_request = router._dispike_instance.view_event_function_return_type(
            _event_name, EventTypes.COMMAND
        )
        _type_hinted_returned_value = _type_hinted_request["return"]
        if _type_hinted_returned_value == DiscordResponse:
            _get_res = await router._dispike_instance.emit(
                _event_name, EventTypes.COMMAND, **arguments
            )

            logger.debug(_get_res.response)
            return _get_res.response
        elif (
            _type_hinted_returned_value == DeferredResponse
            or _type_hinted_returned_value == DeferredEmphericalResponse
        ):
            logger.debug("This is a deferred response...")

            asyncio.create_task(
                _run_and_log_async(
                    router._dispike_instance.emit(
                        _event_name, EventTypes.COMMAND, **arguments
                    )
                )
            )
            return _type_hinted_returned_value.response

        elif _type_hinted_returned_value == dict:
            return await router._dispike_instance.emit(
                _event_name, EventTypes.COMMAND, **arguments
            )
    except KeyError:
        logger.error(
            "unable to find return value for type hint.. resorting to guessing.."
        )
        if _RAISE_FOR_TESTING:
            raise AssertionError("No hinting!")  # pragma: no cover
    except Exception:
        logger.exception("unhandled exception for returning hinted value")
        raise

    interaction_data = await router._dispike_instance.emit(
        _event_name, EventTypes.COMMAND, **arguments
    )
    if isinstance(interaction_data, DiscordResponse):
        interaction_data: DiscordResponse
        return interaction_data.response

    if isinstance(interaction_data, dict):
        return interaction_data

    warnings.warn(
        f"Command {_event_name} has not been configured with a valid callback function.",
        UserWarning,
    )
    logger.warning(
        f"Command {_event_name} has not been configured with a valid callback function."
    )

    # Backup response, simply acknowledge. (Type 5)
    return DiscordResponse(
        content="**Warning**: This command has not been configured. (sent by dispike)",
        empherical=True,
    )



from pydantic import BaseModel
from enum import Enum


class ChannelTypes(int, Enum):
    GUILD_TEXT = 0  # a text channel within a server
    DM = 1  # a direct message between users
    GUILD_VOICE = 2  # a voice channel within a server
    GROUP_DM = 3  # a direct message between multiple users
    GUILD_CATEGORY = 4  # an organizational category that contains up to 50 channels
    GUILD_NEWS = (
        5  # a channel that users can follow and crosspost into their own server
    )
    GUILD_STORE = 6  # a channel in which game developers can sell their game on Discord
    GUILD_NEWS_THREAD = 10  # a temporary sub-channel within a GUILD_NEWS channel
    GUILD_PUBLIC_THREAD = 11  # a temporary sub-channel within a GUILD_TEXT channel
    GUILD_PRIVATE_THREAD = 12  # a temporary sub-channel within a GUILD_TEXT channel that is only viewable by those invited and those with the MANAGE_THREADS permission
    GUILD_STAGE_VOICE = 13  # a voice channel for hosting events with an audience


class PartialChannel(BaseModel):
    """A partial representation for a discord channel. This is found in a Application Command Interaction Data Resolved Structure

    Partial Channel objects only have id, name, type and permissions attributes.

    """

    class Config:
        arbitary_types_allowed = True

    id: int
    type: ChannelTypes
    name: str
    permissions: str



import typing
from enum import Enum

from pydantic import BaseModel

from dispike.errors.components import (
    InvalidComponentError,
    ComponentCombinationError,
    SelectMenuOptionError,
)


class ComponentTypes(int, Enum):

    """Easy access to component types.

    Attributes:
        ACTION_ROW (int): Represents Type 1
        BUTTON (int): Represents Type 2
        SELECT_MENU (int): Represents Type 3
    """

    ACTION_ROW = 1
    BUTTON = 2
    SELECT_MENU = 3


class ButtonStyles(int, Enum):

    """Easy access to button styles.

    Attributes:
        PRIMARY (int): Represents Style 1
        SECONDARY (int): Represents Style 2
        SUCCESS (int): Represents Style 3
        DANGER (int): Represents Style 4
        LINK (int): Represents Style 5
    """

    PRIMARY = 1
    SECONDARY = 2
    SUCCESS = 3
    DANGER = 4
    LINK = 5


class PartialEmoji:
    """Represents a partial Discord emoji"""

    def __init__(self, name: str, id: str = None, animated: bool = None):
        """
        Args:
            name (str): The emojis name, or unicode symbol.
            id (str, optional): id of the emoji, if not a unicode one.
            animated (bool, options): bool Specifying if this emoji should be animated, if not a unicode one.
        """
        self.name = name
        self.id = id
        self.animated = animated

    def to_dict(self):
        _dict = {"name": self.name}
        if self.id:
            _dict["id"] = self.id
        if self.animated is not None:
            _dict["aniamted"] = self.animated
        return _dict


class Button:
    """Represents a partial Discord emoji"""

    def __init__(
        self,
        label: str = None,
        custom_id: str = None,
        disabled: bool = False,
        style: ButtonStyles = ButtonStyles.PRIMARY,
        emoji: PartialEmoji = None,
    ):
        """
        Args:
            label (str): label of the button.
            custom_id (str, optional): custom id of the button.
            disabled (bool, optional): bool specifying if this button should be disabled.
            style (ButtonStyles, optional): style of the button. Cannot be LINK
            emoji (PartialEmoji, optional): a partial emoji.
        """
        self.type = ComponentTypes.BUTTON

        if style == ButtonStyles.LINK:
            raise TypeError("style cannot be type Link")
        if not label:
            raise TypeError("label cannot be None")
        if not custom_id:
            raise TypeError("custom_id cannot be None")

        self.style = style
        self.label = label
        self.custom_id = custom_id
        self.disabled = disabled
        self.emoji = emoji

    def to_dict(self):
        temp_dict = {
            "type": self.type.value,
            "style": self.style.value,
            "label": self.label,
            "custom_id": self.custom_id,
        }

        # an emoji is optional
        if self.emoji:
            temp_dict["emoji"] = self.emoji.to_dict()

        return temp_dict


class LinkButton:
    """Represents a partial Discord emoji"""

    def __init__(self, label: str = None, url: str = None, disabled: bool = False):
        """
        Args:
            label (str): Label of the button.
            url (str): Url of the link button.
            disabled (bool, optional): bool Specifying if this button should be disabled.
        """
        self.type = ComponentTypes.BUTTON
        self.style = ButtonStyles.LINK

        if not url:
            raise TypeError("url cannot be None")
        if not label:
            raise TypeError("label cannot be None")

        self.label = label
        self.url = url
        self.disabled = disabled

    def to_dict(self):
        return {
            "type": self.type.value,
            "style": self.style.value,
            "label": self.label,
            "url": self.url,
            "disabled": self.disabled,
        }


class SelectMenu:
    """Represents a Discord select menu"""

    class SelectMenuOption:
        """Represents a Discord select menu option"""

        def __init__(
            self,
            label: str = None,
            value: str = None,
            description: str = None,
            emoji: PartialEmoji = None,
            default: bool = False,
        ):
            """
            Args:
                label (str): Label of the option.
                value (str): Internal value of the option.
                description (str): Description of the option.
                emoji (PartialEmoji, optional): A partial emoji
                default (bool): Whether or not this option should be the default.
            """

            if not label:
                raise TypeError("label cannot be None")
            if not value:
                raise TypeError("value cannot be None")

            self.label = label
            self.value = value
            self.description = description
            self.emoji = emoji
            self.default = default

        def to_dict(self):
            temp_dict = {
                "label": self.label,
                "value": self.value,
                "default": self.default,
            }

            if self.description:
                temp_dict["description"] = self.description
            if self.emoji:
                temp_dict["emoji"] = self.emoji.to_dict()

            return temp_dict

    def __init__(
        self,
        custom_id: str,
        options: typing.List[SelectMenuOption],
        placeholder: str,
        min_values: int,
        max_values: int,
        disabled: False,
    ):
        """
        Args:
            custom_id (str): Custom id of the menu.
            options (SelectMenu.SelectMenuOption): List of SelectMenu.SelectMenuOption's
            placeholder (str, optional): Text that's shown before a user selects anything.
            min_values (int): The minimum amount of values a user can select. Cannot be lower than 0
            max_values (int): The maximum amount of values a user can select. Cannot be higher than the amount of options.
            disabled (bool, optional): Whether to disable menu
        """
        self.type = ComponentTypes.SELECT_MENU
        self.options = options

        for option in options:
            if not isinstance(option, SelectMenu.SelectMenuOption):
                raise SelectMenuOptionError(f"{option} is not type SelectMenuOption")

        if not placeholder:
            raise TypeError("placeholder cannot be None")
        if not custom_id:
            raise TypeError("custom_id cannot be None")

        self.placeholder = placeholder
        self.custom_id = custom_id

        if not min_values or not max_values:
            raise TypeError("min_values or max_values must not be None")

        if max_values > len(options):
            raise TypeError(
                "max_values cannot be larger than how many options you have"
            )
        if min_values < 1:
            raise TypeError("min_values cannot be less than 0")

        self.min_values = min_values
        self.max_values = max_values

        self.disabled = disabled

    def to_dict(self):
        return {
            "type": self.type.value,
            "options": [x.to_dict() for x in self.options],
            "custom_id": self.custom_id,
            "min_values": self.min_values,
            "max_values": self.max_values,
            "placeholder": self.placeholder,
            "disabled": self.disabled,
        }




class ActionRow:
    """Represents a action row component"""

    def __init__(
        self, components: typing.List[typing.Union[Button, LinkButton, SelectMenu]]
    ):
        """
        Args:
            components (typing.List[typing.Union[Button, LinkButton, SelectMenu]]): Components for this action row. You cannot combine buttons and select menus
        """

        self.type = ComponentTypes.ACTION_ROW

        # Do a heap of stuff to figure out if a select menu is being combined with a button
        contains_button = False
        contains_select_menu = False
        for component in components:
            if isinstance(component, (Button, LinkButton)):
                contains_button = True
            elif isinstance(component, SelectMenu):
                contains_select_menu = True
            else:
                # raise error if neither.
                raise InvalidComponentError(type(component))

        if contains_button and contains_select_menu:
            raise ComponentCombinationError(
                "You cannot combine buttons and select menus"
            )

        self.components = components

    def to_dict(self):
        return {
            "type": self.type.value,
            "components": [x.to_dict() for x in self.components],
        }

    def __eq__(self, other: "ActionRow"):
        return (
            self.__dict__ == other.__dict__
            and [x.to_dict() for x in self.components]
            == [x.to_dict() for x in self.components]
            and self.__class__ == other.__class__
        )



from pydantic import BaseModel

from datetime import datetime

import typing

try:
    from typing import Literal  # pragma: no cover
except ImportError:  # pragma: no cover
    # backport
    from typing_extensions import Literal  # pragma: no cover

from .user import User


class Member(BaseModel):

    """A representation of a discord member, intended for you to easily access attributes. this is not intended for you to edit, and will not
    be accepted as an argument in any function.
    """

    class Config:
        arbitrary_types_allowed = True

    user: User
    roles: typing.List[str]
    premium_since: typing.Union[
        None, str, int
    ]  # ? Discord docs doesn't talk about this so..
    permissions: str
    pending: bool
    nick: typing.Union[None, str]
    mute: typing.Optional[bool] = None
    joined_at: datetime
    deaf: typing.Optional[bool] = None


class PartialMember(BaseModel):
    """A partial representation for a discord user. This is found in a Application Command Interaction Data Resolved Structure

    Partial Member objects are missing user, deaf and mute attributes.

    """

    class Config:
        arbitrary_types_allowed = True

    roles: typing.List[str]
    premium_since: typing.Union[
        None, str, int
    ]  # ? Discord docs doesn't talk about this so..
    permissions: str
    pending: bool
    nick: typing.Union[None, str]
    joined_at: datetime



from .helper.embed import Embed
from .creating.components import ActionRow
import typing
from .errors.network import DiscordAPIError
from .errors.response import ContentIsEmpty, InvalidDiscordResponse
from loguru import logger
import warnings

try:
    from typing import Literal  # pragma: no cover
except ImportError:  # pragma: no cover
    # backport
    from typing_extensions import Literal  # pragma: no cover

if typing.TYPE_CHECKING:
    from .main import Dispike  # pragma: no cover
    from .creating import IncomingDiscordInteraction  # pragma: no cover
    from .creating.allowed_mentions import AllowedMentions  # pragma: no cover


class DiscordResponse(object):
    """Represents an outgoing Discord Response

    Attributes:
        content (str): A plain-text response to a user
        tts (bool): bool returning if the message should be spoken via tts.
        embeds (dict): a List representing .to_dict of an Embed object.
        response (dict): a valid response represented in a dict, to later be converted to JSON.
    """

    def __init__(
        self,
        content: str = None,
        tts: bool = False,
        embeds: typing.List[Embed] = [],
        show_user_input: bool = False,
        follow_up_message=False,
        empherical=False,
        allowed_mentions: "AllowedMentions" = None,
        action_row: ActionRow = None,
        update_message=False,
    ):
        """Initialize a DiscordResponse, you can either pass data into here, or
        simply create a DiscordResponse() and edit via properties.

        Args:
            content (str, optional): A plain-text response to a user
            tts (bool, optional): bool returning if the message should be spoken via tts
            embeds (typing.List[Embed], optional): a List representing .to_dict of an Embed object
            show_user_input (bool, optional): Whether to delete the user's message of calling the command after responding.
            follow_up_message (bool, optional): Whether this is a follow up to a previous message.
            empherical (bool, optional): Whether to send message as an empherical message.
            allowed_mentions (List[AllowedMentions], optional): Let discord filter mentions per configuration.
            update_message (bool, optional): Whether to edit the original message this is responding too.
        """
        if content is not None:
            if not isinstance(content, str):
                raise TypeError(f"Content must be a string")
            elif content == "":
                content = None

        # if isinstance(content, str) == False or content == "" or content != None:
        #    raise TypeError(f"content must be a string. recieved: {content}")

        if not isinstance(tts, bool):
            raise TypeError("tts must be a bool")

        self._content = content
        self._tts = tts
        self._embeds = embeds

        if action_row:
            self._action_row = action_row
        else:
            self._action_row = None

        if show_user_input:
            # show deprecation warning
            warnings.warn(
                "show_user_input is deprecated by Dispike as it is no longer supported by Discord. Future versions of dispike may remove this parameter entirely.",
                DeprecationWarning,
            )  # pragma: no cover
            logger.warning(
                "show_user_input is deprecated by Dispike as it is no longer supported by Discord. Future versions of dispike may remove this parameter entirely."
            )  # pragma: no cover

        self._type_response = 4

        if update_message:
            self._type_response = 7

        self._is_followup = follow_up_message
        self._is_empherical = empherical
        self._allowed_mentions = allowed_mentions

    @property
    def embeds(self) -> typing.List[Embed]:
        """Returns a list of embeds to send to.

        Returns:
            List[Embed]: List of embeds in this object.
        """
        return self._embeds

    @property
    def action_row(self) -> ActionRow:
        """Returns a action row.

        Returns:
            ActionRow: The action row.
        """
        return self._action_row

    def set_type_response(self, type: int):
        self._type_response = type

    def add_new_embed(self, embed_to_add: Embed):
        """Append a new embed, provided with a proper Embed object

        Args:
            embed_to_add (Embed): Proper Embed Object

        Raises:
            TypeError: Raised if you do not pass a proper Embed object.
        """
        if isinstance(embed_to_add, Embed):
            self._embeds.append(embed_to_add)
        else:
            raise TypeError("embed must be a Embed object.")

    @property
    def content(self) -> str:
        """Either set or view the plain-text response to the user.

        Returns:
            str: Content provided
        """
        return self._content

    @content.setter
    def content(self, new_content_string: str):
        self._content = str(new_content_string)

    @property
    def tts(self) -> bool:
        """Either set or view the tts attribute for the user.

        Returns:
            bool: tts
        """
        return self._tts

    @tts.setter
    def tts(self, new_tts: bool):
        self._tts = new_tts

    @property
    def response(self) -> dict:
        """A generated valid discord response

        Returns:
            dict: a valid discord response.
        """

        # Watch this, this may cause an issue.
        self.content = "" if self.content is None else self.content

        if self._is_followup:
            _req = {"content": self.content, "data": {}}

            if self.embeds:
                _req["embeds"] = [x.to_dict() for x in self.embeds]

            if self.tts:
                _req["tts"] = True

            if self._is_empherical:
                logger.info("setting empherical")
                _req["flags"] = 1 << 6

            if self._action_row:
                _req["data"]["components"] = [self.action_row.to_dict()]

            return _req

        _req = {
            "type": self._type_response,
            "data": {
                "tts": self.tts,
                "content": self.content,
                "embeds": [x.to_dict() for x in self.embeds],
            },
        }
        if self._is_empherical:
            _req["data"]["flags"] = 1 << 6

        if self._action_row:
            if self.content == None or self.embeds == [] or self.content == "":
                raise InvalidDiscordResponse(
                    "If creating a response with an action row, content or an embed must be present!"
                )

            _req["data"]["components"] = [self.action_row.to_dict()]

        if self._allowed_mentions:
            _req["allowed_mentions"] = self._allowed_mentions.dict()

        return _req

    def _switch_to_followup_message(self):
        self._is_followup = True

    def __call__(self) -> dict:
        return self.response


class DeferredResponse:
    """Use this class to hint to dispike that the response will be deferred."""

    response = {"type": 5}


class DeferredEmphericalResponse:
    """Use this class to hint to dispike that the response will be deferred and empherical."""

    response = {"type": 5, "data": {"flags": 1 << 6}}


class AcknowledgeComponentResponse:
    """Represents an interaction ACK response"""

    response = {"type": 4, "data": {}}



from pydantic import BaseModel
import typing

try:
    from typing import Literal  # pragma: no cover
except ImportError:  # pragma: no cover
    # backport
    from typing_extensions import Literal  # pragma: no cover


class User(BaseModel):

    """A representation of a User object from discord. this is not intended for you to edit, and will not
    be accepted as an argument in any function.
    """

    class Config:
        arbitary_types_allowed = True

    id: int
    username: str
    avatar: typing.Union[None, str]
    discriminator: str
    public_flags: int



from dispike.incoming import (
    IncomingDiscordSlashInteraction,
    IncomingDiscordSlashData,
    SubcommandIncomingDiscordOptionList,
    IncomingDiscordOption,
    SubcommandIncomingDiscordOptionListChild,
)
import typing
from loguru import logger


def determine_event_information(
    interaction: IncomingDiscordSlashInteraction,
) -> typing.Tuple[str, dict]:

    if isinstance(interaction, IncomingDiscordSlashInteraction):

        if interaction.data.options is None:
            return interaction.data.name, {}

        if isinstance(interaction.data.options[0], SubcommandIncomingDiscordOptionList):
            # subcommand event names will be must be xxx.xxx
            _sub_command_arguments = {}

            if hasattr(interaction.data.options[0], "options"):
                # gosh this is hella ugly, i need to rewrite this.
                _event_name = (
                    f"{interaction.data.name}.{interaction.data.options[0].name}"
                )
                if isinstance(
                    interaction.data.options[0].options[0],
                    SubcommandIncomingDiscordOptionListChild,
                ):
                    _event_name += "." + interaction.data.options[0].options[0].name
            # I don't think below will ever happen, but just in case.
            else:  # pragma: no cover
                _event_name = (
                    f"{interaction.data.name}.{interaction.data.options[0].name}"
                )
            for argument in interaction.data.options[0].options:

                if isinstance(argument, SubcommandIncomingDiscordOptionListChild):
                    for sub_child_argument in argument.options:
                        _sub_command_arguments[
                            sub_child_argument.name
                        ] = sub_child_argument.value
                    break
                else:
                    _sub_command_arguments[argument.name] = argument.value

            return _event_name, _sub_command_arguments

        elif isinstance(interaction.data.options[0], IncomingDiscordOption):
            _command_arguments = {}

            for argument in interaction.data.options:
                argument: IncomingDiscordOption
                _command_arguments[argument.name] = argument.value

            return interaction.data.name, _command_arguments

    raise TypeError("unable to determine eventname.")



__version__ = "1.0.1-beta.0"

from .main import Dispike
from .incoming import IncomingDiscordSlashInteraction, IncomingApplicationCommand
from .response import DiscordResponse
from .creating.models import DiscordCommand
from .interactions import PerCommandRegistrationSettings, EventCollection

from loguru import logger

_add_networking_logging_level = logger.level("NETWORK", no=20, color="<fg #C954E2>")



import typing_extensions, typing
from dispike.creating.models import permissions


class PermissionGenerator(object):
    """A helper that helps you create proper permissions for a slash command."""

    def __init__(
        self,
        restrict_to_role: bool = False,
        restrict_to_user: bool = False,
        restricted_role_id: typing.Union[list[int], int] = [],
        restricted_user_id: typing.Union[list[int], int] = [],
        allow_all_except_or_deny_all_except: typing.Union[
            typing_extensions.Literal["allow_all"],
            typing_extensions.Literal["deny_all"],
        ] = "allow_all",
    ):
        """A Helper that helps you create proper permissions for a slash command. Generates a NewApplicationPermission.

        Args:
            restrict_to_role (bool, optional): Whether to restrict to a specific Discord Role ID.. Defaults to False.
            restrict_to_user (bool, optional): Whether to restrict to a specific Discord User ID. Defaults to False.
            restricted_role_id (int, optional): The restricted role ID. Defaults to None.
            restricted_user_id (int, optional): The restricted user ID. Defaults to None.
            allow_all_except_or_deny_all_except (Union["allow_all", "deny_all"], optional): Whether to allow only X or allow everyone else other than X. Defaults to "allow_all".

        Raises:
            TypeError: If `allow_all_except_or_deny_all_except` is not a valid value.
            ValueError: If you attempt to combine restrictions.
        """

        self.__restrict_to_role = restrict_to_role
        self.__restrict_to_user = restrict_to_user

        if restrict_to_role == False:
            self.__restricted_role_id = []
        elif isinstance(restrict_to_role, int):
            self.__restricted_role_id = [restricted_role_id]
        elif isinstance(restrict_to_role, list):
            if restrict_to_role == []:
                self.__restricted_role_id = []
            else:
                for _item in restricted_role_id:
                    if isinstance(_item, int) == False:
                        raise ValueError(
                            f"Value passed to restricted roles is not a valid type. Recieved {type(restrict_to_role)}.. Need <int>!"
                        )
                self.__restricted_role_id = restricted_role_id
        else:
            raise TypeError(
                "Unknown type recieved for role.. If this is a string, go ahead and convert it to an int"
            )

        if restrict_to_user == False:
            self.__restricted_user_id = []
        elif isinstance(restrict_to_user, int):
            self.__restricted_user_id = [restricted_user_id]
        elif isinstance(restrict_to_user, list):
            if restricted_user_id == []:
                self.__restricted_user_id = []
            else:
                for _item in restricted_role_id:
                    if isinstance(_item, int) == False:
                        raise ValueError(
                            f"Value passed to restricted users is not a valid type. Recieved {type(restrict_to_user)}.. Need <int>!"
                        )
                self.__restricted_user_id = restricted_user_id
        else:
            raise TypeError(
                "Unknown type recieved for user.. If this is a string, go ahead and convert it to an int"
            )

        if allow_all_except_or_deny_all_except not in ["allow_all", "deny_all"]:
            raise TypeError(
                f"{allow_all_except_or_deny_all_except} is unknown. Cannot determine whether to deny or allow based on permission"
            )
        else:
            if allow_all_except_or_deny_all_except == "deny_all":
                self.__allow_all_except_or_deny_all_except = False
            else:
                self.__allow_all_except_or_deny_all_except = True

        if self.__restrict_to_role and self.__restricted_role_id is []:
            raise ValueError("Cannot restrict to role without a role id")

        if self.__restrict_to_user and self.__restricted_user_id is []:
            raise ValueError("Cannot restrict to user without a user id")

    @property
    def restricted_to_role(self) -> bool:
        return self.__restrict_to_role

    @property
    def restricted_to_user(self) -> bool:
        return self.__restrict_to_user

    @property
    def restricted_role_id(self) -> typing.List[int]:
        return self.__restricted_role_id

    @property
    def restricted_user_id(self) -> typing.List[int]:
        return self.__restricted_user_id

    @property
    def allow_all_except_or_deny_all_except(self) -> bool:
        return self.__allow_all_except_or_deny_all_except

    def remove_role_from_restricted_roles(self, role_id: int):
        if role_id in self.__restricted_role_id:
            self.__restricted_role_id.remove(role_id)

    def add_role_to_restricted_roles(
        self, role_id: int, ignore_if_already_present: bool = False
    ):
        if role_id in self.__restricted_role_id:
            if ignore_if_already_present:
                return
            else:
                raise ValueError(
                    f"Role {role_id} is already in the restricted roles. Cannot add it again."
                )
        else:
            self.__restricted_role_id.append(role_id)

    def remove_user_from_restricted_roles(self, user_id: int):
        if user_id in self.__restrict_to_user:
            self.__restrict_to_user.remove(user_id)

    def add_user_to_restricted_users(
        self, user_id: int, ignore_if_already_present: bool = False
    ):
        if user_id in self.__restrict_to_user:
            if ignore_if_already_present:
                return
            else:
                raise ValueError(
                    f"User {user_id} is already in the restricted users. Cannot add it again."
                )
        else:
            self.__restrict_to_user.append(user_id)

    @allow_all_except_or_deny_all_except.setter
    def set_allow_all_except_or_deny_all_except(self, new_value):
        if new_value not in ["allow_all", "deny_all"]:
            raise TypeError(
                f"{new_value} is unknown. Cannot determine whether to deny or allow based on permission"
            )
        else:
            if new_value == "deny_all":
                self.__allow_all_except_or_deny_all_except = False
            else:
                self.__allow_all_except_or_deny_all_except = True

    @property
    def created(self) -> permissions.NewApplicationPermission:
        """Returns a NewApplicationPermission object based on the current attributes.

        Returns:
            permissions.NewApplicationPermission: A properly configured object.
        """

        _permissions_generated = []
        if self.restricted_to_role:
            if self.__restricted_role_id == []:
                raise ValueError("Cannot create restriction for roles with no role ids")

            for role_to_restrict in self.restricted_role_id:
                _permissions_generated.append(
                    permissions.ApplicationCommandPermissions(
                        id=role_to_restrict,
                        type=permissions.ApplicationCommandPermissionType.ROLE,
                        permission=self.__allow_all_except_or_deny_all_except,
                    )
                )

        if self.restricted_to_user:
            if self.__restricted_user_id == []:
                raise ValueError("Cannot create restriction for roles with no role ids")

            for role_to_restrict in self.restricted_user_id:
                _permissions_generated.append(
                    permissions.ApplicationCommandPermissions(
                        id=role_to_restrict,
                        type=permissions.ApplicationCommandPermissionType.USER,
                        permission=self.__allow_all_except_or_deny_all_except,
                    )
                )

        return permissions.NewApplicationPermission(permissions=_permissions_generated)



import dataclasses
import typing

from pydantic import BaseModel, Extra, validator, root_validator
from pydantic.error_wrappers import ValidationError
from pydantic.errors import ArbitraryTypeError
from enum import Enum

try:
    from typing import Literal  # pragma: no cover
except ImportError:  # pragma: no cover
    # backport
    from typing_extensions import Literal  # pragma: no cover

if typing.TYPE_CHECKING:  # pragma: no cover
    static_check_init_args = dataclasses.dataclass
else:

    def static_check_init_args(cls):
        return cls


class OptionTypes(int, Enum):
    """Easy access to option types.

    Attributes:
        BOOLEAN (int): Represents Type 5
        CHANNEL (int): Represents Type 7
        INTEGER (int): Represents Type 4
        ROLE (int): Represents Type 8
        STRING (int): Represents Type 3
        SUB_COMMAND (int): Represents Type 1
        SUB_COMMAND_GROUP (int): Represents Type 2
        USER (int): Represents Type 6
        MENTIONABLE (int): Represents Type 9
        NUMBER (int): Represents Type 10
    """

    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10


class CommandTypes(int, Enum):
    """Easy access to command types.

    Attributes:
        SLASH (int): Represents Type 1
        USER (int): Represents Type 2
        MESSAGE (int): Represents Type 3
    """

    SLASH = 1
    USER = 2
    MESSAGE = 3


@static_check_init_args
class CommandChoice(BaseModel):

    """Represents a key-value command choice.

    Attributes:
        name (str): Name of the choice.
        value (str): Value of the choice.
    """

    name: str
    value: typing.Union[str, int, float]


@static_check_init_args
class CommandOption(BaseModel):

    """Represents a standard command option (not a subcommand).

    Attributes:
        name (str): Name of the option.
        description (str): Description of the option.
        type (OptionTypes): The option type.
        required (bool): Whether or not this option is required.
        choices (typing.Union[typing.List[dict], typing.List[CommandChoice]], optional): Possible choices for this option for the user to pick from.
        options (typing.Union[typing.List[CommandChoice], typing.List], optional): If the option is a subcommand or subcommand group type, this nested options will be the parameters.
    """

    class Config:
        arbitrary_types_allowed = True

    name: str
    description: str
    type: OptionTypes
    required: bool = False
    choices: typing.Optional[
        typing.Union[typing.List[dict], typing.List[CommandChoice]]
    ] = None

    options: typing.Optional[
        typing.Union[typing.List[CommandChoice], typing.List]
    ] = None

    # @validator("options")
    # def options_allow_only_if_subcommand(cls, v):
    #   if cls.type != 1:
    #       raise ValidationError("Type must be 1 in order to have options.")
    #   return v


@static_check_init_args
class SubcommandOption(BaseModel):

    """Represents a subcommand group usually you would put this as an option in a DiscordCommand

    Attributes:
        name (str): Name of this group.
        description (str): Description of this group.
        options (typing.List[CommandOption]): Options for this group.
    """

    class Config:
        arbitrary_types_allowed = True

    name: str
    description: str
    options: typing.List[CommandOption]
    type: Literal[2] = 2

    # @validator("options", pre=True, always=True)
    # def options_must_contain_type_1(cls, v):  # pylint: disable=no-self-argument
    #    item: CommandOption
    #    for item_location, item in enumerate(v):
    #        if isinstance(item, CommandOption):
    #            _type = item.type
    #            _error_name = item.name
    #        elif isinstance(item, dict):
    #            _type = item["type"]
    #            _error_name = item["name"]
    #
    #        if _type != 1:
    #            raise ValueError(
    #                f"CommandOptions <{_error_name}> located <{item_location}> must be have type of 1 due to parent being a subcommand."
    #            )
    #    return v


@static_check_init_args
class DiscordCommand(BaseModel):

    """Represents a discord command.

    Attributes:
        id (int, optional): Id of this command.
        name (str): Name of this command.
        description (str): Description of this command.
        options (typing.List[typing.Union[SubcommandOption, CommandOption]]): Options for this command.
        default_permission (boolean): whether the command is enabled by default when the app is added to a guild. Defaults to True.
        type (Union[CommandTypes, int]): The type of command. This defaults to SLASH
    """

    id: typing.Optional[int]
    name: str
    description: str = ""
    options: typing.List[typing.Union[SubcommandOption, CommandOption]] = []
    default_permission: bool = True
    type: typing.Union[CommandTypes, int] = CommandTypes.SLASH

    @root_validator
    def validate_correct_type_requirements(cls, values):
        _cls_type, _cls_description, _cls_options = (
            values.get("type"),
            values.get("description"),
            values.get("options"),
        )
        if _cls_type == CommandTypes.SLASH:
            if _cls_description == "":
                raise ValueError("Slash commands require a description")
        if _cls_type == CommandTypes.MESSAGE or _cls_type == CommandTypes.USER:
            if _cls_description != "":
                raise ValueError("Context commands cannot have a description")
            if _cls_options:
                raise ValueError("Context commands cannot have options")

        return values

    @validator("type", pre=True, always=True)
    def set_value_type(cls, v):
        if isinstance(v, CommandTypes):
            return v.value
        return v



from pydantic import BaseModel
import typing


"""
class RoleTags(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    bot_id: typing.Optional[typing.Any] = None
    intergration_id: typing.Optional[typing.Any] = None
    permium_subscriber = typing.Optional[typing.Any] = None
"""


class Role(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    id: int
    name: str
    color: str
    hoist: bool
    position: int
    permissions: str
    managed: str
    mentionable: bool
    tags: typing.Optional[typing.Dict] = {}


from dispike.eventer import EventTypes
from dispike.errors.events import InvalidEventType
from loguru import logger
import inspect

import typing

if typing.TYPE_CHECKING:
    from .main import Dispike  # pragma: no cover
    from dispike.creating.models.options import DiscordCommand  # pragma: no cover


def on(
    event: str,
    type: EventTypes = EventTypes.COMMAND,
    func: typing.Callable = None,
):
    """A proxy decorator for registering commands. This decorator will add a number of attributes within the object.
    You can use this decorator to register commands with ``.register_command`` function on the main Dispike object.

    Args:
        event (str): Event callback name, usually the discord command name.
        type (EventTypes, optional): The event type. Defaults to EventTypes.COMMAND.
        func (typing.Callable, optional): Function, will automatically be added if a decorator..

    Raises:
        InvalidEventType: If invalid EventType. Will not raise if it's a string.

    Returns:
        typing.Callable: A function with the attributes added.
    """
    if not isinstance(type, EventTypes):
        if isinstance(type, str):  # pragma: no cover
            logger.warning(
                "Passing a unknown EventType, this may cause issues and is unsupported"
            )  # noqa
        else:
            # TODO: Maybe it's not good to overrwrite a default python function. Maybe change type to a different value?
            raise InvalidEventType(type)  # pragma: no cover

    def on(func):
        if not inspect.iscoroutinefunction(func):
            pass
            # raise TypeError("Function must be a async function.")
        func._dispike_event_type = type
        func._dispike_event_name = event
        return func

    return on(func) if func else on


class PerCommandRegistrationSettings(object):
    """Allows for per-basis configuration of a command."""

    def __init__(self, schema: "DiscordCommand", guild_id: str = None):
        self.schema = schema
        self.guild_id = guild_id


class EventCollection(object):
    """Base object to inherit to help developers keep similar commands in a group."""

    @staticmethod
    def command_schemas() -> typing.List[
        typing.Union[PerCommandRegistrationSettings, "DiscordCommand"]
    ]:
        """Function to return list of DiscordCommands or PerCommandRegistrationSettings objects.

            Returns:
                List[
            typing.Union[PerCommandRegistrationSettings, "DiscordCommand"]
        ]: ...
        """
        return []

    @staticmethod
    def registered_commands() -> typing.List[typing.Callable]:
        """Function to return list of event callbacks. Usually not used, Dispike will automatically find functions with the interactions decorator.

        Returns:
            typing.List[typing.Callable]: Functions.
        """
        return []



from pydantic import BaseModel

import typing

try:
    from typing import Literal  # pragma: no cover
except ImportError:  # pragma: no cover
    # backport
    from typing_extensions import Literal  # pragma: no cover

from .user import User


class MessageAttachment(BaseModel):
    """
    A message attachment
    """

    id: str
    filename: str
    content_type: str
    size: int
    url: str
    proxy_url: str
    height: int = None
    width: int = None


class Message(BaseModel):

    """A representation of a discord message, intended for you to easily access attributes. this is not intended for you to edit, and will not
    be accepted as an argument in any function.
    """

    class Config:
        arbitrary_types_allowed = True

    # TODO: Make this a lot better, there is lots of unimplemented stuff here...

    attachments: typing.List[MessageAttachment]
    author: User
    channel_id: str
    # components:
    content: str
    edited_timestamp: str = None
    # embeds
    flags: int
    id: int
    mention_everyone: bool
    # mention_roles:
    # mentsions
    pinned: bool
    timestamp: str
    tts: bool
    type: int



import typing
from ..discord_types.member import PartialMember
from ..discord_types.channel import PartialChannel
from ..discord_types.user import User
from ..discord_types.role import Role

from loguru import logger

if typing.TYPE_CHECKING: # pragma: no cover
    from ..incoming_interactions import IncomingDiscordSlashInteraction 


_class_return = {
    "members": PartialMember,
    "channels": PartialChannel,
    "users": User,
    "roles": Role,
}


@logger.catch(reraise=True, message="Unable to lookup query for resolve")
def resolved_interactions_finder(
    cls: "IncomingDiscordSlashInteraction",
    query: typing.Union[str, int],
    type_to_determine: str,
) -> typing.Union[PartialMember, None]:
    """
    Finds all interactions that are resolved.

    :param query: The query to search for.
    :param type_to_determine: The type of interaction to search for.
    :return: A list of all interactions that are resolved.
    """
    _grab_member = cls.data.resolved[type_to_determine].get(query, None)
    if _grab_member is not None:
        return _class_return[type_to_determine](**_grab_member)
    return _grab_member


def lookup_resolved_member_helper(cls, member_id: str) -> PartialMember:
    return resolved_interactions_finder(cls, str(member_id), "members")


def lookup_resolved_channel_helper(cls, channel_id: str) -> PartialChannel:
    return resolved_interactions_finder(cls, str(channel_id), "channels")


def lookup_resolved_user_helper(cls, user_id: str) -> User:
    return resolved_interactions_finder(cls, str(user_id), "users")


def lookup_resolved_role_helper(cls, role_id: str) -> Role:
    return resolved_interactions_finder(cls, str(role_id), "roles")


class InvalidComponentError(Exception):
    def __init__(self, type):
        self.message = f"{type} is not a valid component class"

    def __str__(self):
        return self.message


class ComponentCombinationError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class SelectMenuOptionError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message



from pydantic import BaseModel
from enum import Enum
import typing
import dataclasses

if typing.TYPE_CHECKING:  # pragma: no cover
    static_check_init_args = dataclasses.dataclass
else:

    def static_check_init_args(cls):
        return cls


@static_check_init_args
class ApplicationCommandPermissionType(int, Enum):
    """The permission type of the Application Command."""

    ROLE = 1
    USER = 2


@static_check_init_args
class ApplicationCommandPermissions(BaseModel):
    """An individual permission for a command. """

    id: int
    type: ApplicationCommandPermissionType
    permission: bool


@static_check_init_args
class GuildApplicationCommandPermissions(BaseModel):
    """ Returned when fetching the permissions for a command in a guild. """

    id: int
    application_id: int
    permissions: typing.List[ApplicationCommandPermissions]


@static_check_init_args
class NewApplicationPermission(BaseModel):
    permissions: typing.List[ApplicationCommandPermissions]



class BotTokenNotProvided(Exception):
    """Exception that is raised when you attempt to use a certain
    function, method or class without providing a bot token.
    """

    def __init__(self, functionality_name: str = "unknown"):
        self.func_name = functionality_name

    def __str__(self) -> str:
        return f"This functionality ({self.func_name}) cannot be used without supplying a bot token to Dispike"

    def __repr__(self) -> str:
        return f"This functionality ({self.func_name}) cannot be used without supplying a bot token to Dispike"


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dispike',
 'dispike.creating',
 'dispike.creating.models',
 'dispike.errors',
 'dispike.eventer_helpers',
 'dispike.followup',
 'dispike.followup.storage',
 'dispike.helper',
 'dispike.incoming',
 'dispike.incoming.attribute_helpers',
 'dispike.incoming.discord_types',
 'dispike.middlewares']

package_data = \
{'': ['*']}

install_requires = \
['PyNaCl>=1.4.0,<2.0.0',
 'async-timeout>=3.0.1,<4.0.0',
 'fastapi>=0.63.0,<0.64.0',
 'httpx>=0.16.1,<0.17.0',
 'loguru>=0.5.3,<0.6.0',
 'pydantic==1.8.2',
 'typing-extensions>=3.7.4,<4.0.0',
 'uvicorn>=0.13.2,<0.14.0']

setup_kwargs = {
    'name': 'dispike',
    'version': '1.0.1b0',
    'description': 'An independent, simple to use, powerful framework for creating interaction-based Discord bots. Powered by FastAPI',
    'long_description': '<div align="center">\n<br>\n  <h1> dispike </h1>\n  <i> ⚙️  A simple to use, powerful framework for creating stateless, independent bots using <a href="https://discord.com/developers/docs/interactions/slash-commands"> Discord Slash Commands.</a> </i>\n  <br>\n  <br>\n    <a > ⚡ Powered by <a href="https://github.com/tiangolo/fastapi"> FastAPI.</a> </a>\n  <br>\n  <br>\n  <p align="center">\n    <img src="https://codecov.io/gh/ms7m/dispike/branch/master/graph/badge.svg?token=E5AXLZDP9O">\n    <img src="https://github.com/ms7m/dispike/workflows/Test%20Dispike/badge.svg?branch=master">\n    <img src="https://img.shields.io/badge/Available%20on%20PyPi-Dispike-blue?logo=pypi&link=%22https://pypi.org/project/dispike%22">\n    <img src="https://img.shields.io/badge/dynamic/json?color=blue&label=PyPi%20Version&query=%24.info.version&url=https%3A%2F%2Fpypi.org%2Fpypi%2Fdispike%2Fjson">\n  </p>\n  <br>\n</div>\n\n\n## 📦 Installation\n\n\n**Latest stable-version**\n```\npip install dispike\n```\n\n## 📚 Learn more\n- Read documentation [here](https://dispike.ms7m.me)\n- See an example bot [here](https://github.com/ms7m/dispike-example)\n- Join our Discord Server [here](https://discord.gg/yGgRmEYjju)\n\n***\n<div align="center">\n<h2> 🧑\u200d💻 Quick Start Examples </h2>\n</div>\n\n\n### Basic\n\n```python\n\nfrom dispike import Dispike, DiscordCommand, DiscordResponse\nfrom dispike import IncomingDiscordSlashInteraction\nfrom dispike.helper import Embed\n\nbot = Dispike(...)\n\n\ncommand = DiscordCommand(\n  name="stock", description="Get the latest active stocks in the market!"\n)\n\n\n@bot.on("stock")\nasync def handle_stock_request(stockticker: str, ctx: IncomingDiscordSlashInteraction) -> DiscordResponse:\n  get_price = function(stockticker...)\n  \n  embed=Embed()\n  embed.add_field(name="Stock Price for {stockticker}.", value="Current price is {get_price}", inline=True)\n  embed.set_footer(text="Request received by {ctx.member.user.username}")\n  return DiscordResponse(embed=embed)\n\n\n\nif __name__ == "__main__":\n    bot.register(command)\n    bot.run()\n```\n\n\n### Advanced\n\n```python\nimport dispike\nfrom dispike import interactions, DiscordCommand, DiscordResponse\nfrom dispike import IncomingDiscordSlashInteraction\nfrom dispike.helper import Embed\n\n\nclass SampleGroupCollection(interactions.EventCollection):\n\n    def __init__(self):\n        self._api_key = "..."\n\n    def command_schemas(self):\n        return [\n            DiscordCommand(\n                name="lateststocks", description="Get the highest performing stocks in the market currently!"\n            ),\n            interactions.PerCommandRegistrationSettings(\n                schema=DiscordCommand(\n                    name="price",\n                    description="return ticker price for server",\n                    options=[],\n                ),\n                guild_id=11111111,\n            )\n        ]\n\n    def get_stock_information(self, stock_ticker):\n        return ...\n\n    def get_portfolio_stats(self, user_id):\n        return ...\n\n    @interactions.on("lateststocks")\n    async def latest_stocks(self, ctx: IncomingDiscordSlashInteraction) -> DiscordResponse:\n        embed = Embed()\n\n        # check user\'s porfolio by looking in the database by their discord ID\n        portfolio_stats = self.get_portfolio_stats(\n            ctx.member.user.id\n        )\n\n        embed.add_field(name="Stocks are doing good!", value=f"Current portfolio is {portfolio_stats}", inline=True)\n        embed.set_footer(text="Request received by {ctx.member.user.username}")\n        return DiscordResponse(embeds=[embed])\n\n    @interactions.on("price")\n    async def get_stock_price(self, ctx: IncomingDiscordSlashInteraction, ticker: str) -> DiscordResponse:\n        embed = Embed()\n        embed.add_field(name=f"Stock Price for 1.",\n                        value=f"Current price is {self.get_stock_information(ticker)}", inline=True)\n        embed.set_footer(text="Request received by {ctx.member.user.username}")\n        return DiscordResponse(embeds=[embed])\n\n## Inside seperate file\n\nfrom dispike import Dispike, DiscordCommand\n\nbot = Dispike(...)\n\nbot.register_collection(SampleGroupCollection(), register_command_with_discord=True)\n\nif __name__ == "__main__":\n    bot.run(port=5000)\n```\n\n## Discord API Coverage\n<details><summary>View Coverage</summary>\n<p>\n\n| API Endpoint   |      Implementation   |\n|----------|:-------------:|\n| Get Global Application Commands |  **✅ Implemented** |\n| Create Global Application Command |    **✅ Implemented**   |\n| Edit Global Application Command |  **✅ Implemented** |\n| Delete Global Application Command | **✅ Implemented** |\n| Create Guild Application Command | **✅ Implemented** |\n| Edit Guild Application Command | **✅ Implemented** |\n| Delete Guild Application Command | **✅ Implemented** |\n| Create Interaction Response | **✅ Implemented** |\n| Edit Original Interaction Response | **✅ Implemented**|\n| Delete Original Interaction Response | **✅ Implemented** |\n| Create Followup Message |**✅ Implemented** |\n| Edit Followup Message | **✅ Implemented** |\n| Delete Followup Message | **✅ Implemented** |\n| Data Models and Types | **✅ Implemented** |\n| ApplicationCommand | **✅ Implemented** |\n| ApplicationCommandOption | **✅ Implemented** |\n| ApplicationCommandOptionType | **✅ Implemented** |\n| ApplicationCommandOptionChoice | **✅ Implemented** |\n| Interaction | **✅ Implemented** |\n| Interaction Response | **✅ Implemented** |\n| Message Components | **✅ Implemented** |\n| Buttons (Message Components) | **✅ Implemented** |\n| Action Rows (Message Components) | **✅ Implemented** |\n| Message Select (Message Components) | **✅ Implemented** |\n\n</p>\n</details>\n\n## ℹ️ Notice\n\n- Python 3.6+\n- Does not speak over the discord gateway. [discord-py-slash-command is what you are looking for.](https://github.com/eunwoo1104/discord-py-slash-command). \n- You will need a server to accept connections directly from discord!\n\n\n## 🧑\u200d💻 Development\n\nHelp is wanted in mantaining this library. Please try to direct PRs to the ``dev`` branch, and use black formatting (if possible).\n\n# 🎉 Special Thanks\n- [Squidtoon99](https://github.com/Squidtoon99)\n- [marshmallow](https://github.com/mrshmllow)\n',
    'author': 'Mustafa Mohamed',
    'author_email': 'ms7mohamed@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ms7m/dispike',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)



class InvalidEventType(Exception):
    def __init__(self, event_type_passed):
        self._event_type_value = type(event_type_passed)

    def __str__(self):
        return f"An invalid event type: {self._event_type_value} was passed. An event type must be a string!"



class ContentIsEmpty(Exception):
    def __str__(self) -> str:
        return "No string was passed for content. You must add a text content to a DiscordResponse"

    def __repr__(self) -> str:
        return "No string was passed for content. You must add a text content to a DiscordResponse"


class InvalidDiscordResponse(Exception):
    pass


class NoResolvedInteractions(Exception):
    pass